/*

 compat.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "compat.h"
#include "driver.h"
#include "path.h"
#include "crc8.h"
#include "partition.h"
#include "log.h"
#include "arguments.h"
#include "buffer.h"
#include "ideservd.h"
#include "directory.h"
#include "utf8.h"
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define OPEN_RONLY 0
#define OPEN_STATUS 2
#define OPEN_WONLY 4
#define OPEN_ERR 6

static int frame_error(const char *mode) {
    seterror(ER_FRAME_ERROR, 1);
    log_printf("%s: Frame error", mode);
    return 1;
}

static int crc_error(const char *mode) {
    seterror(ER_CRC_ERROR, 1);
    log_printf("%s: CRC error", mode);
    return 1;
}

static void directory_error(const char *path) {
    errnotochannel15(errno);
    log_printf("Open: Couldn't open the directory \"%s\": %s(%d)", path, strerror(errno), errno);
}

int openfile_compat(const Driver *driver, const Arguments *arguments, unsigned int flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    const int debug = flags & FLAG_DEBUG;
    Petscii cmd[256];
    struct timeval start;
    unsigned char status = OPEN_ERR;
    Petscii mode = 'R';
    unsigned int channel, fp;
    Path outpath;
    Buffer *buffer;
    int r;

    r = driver->getb(1);
    if (usecrc) crc_add_byte(r);
    channel = r & 0x0f;
    buffer = buffers_get(channel);

    fp = 0;
    while (fp < 255) {
        r = driver->getb(0);
        if (r == EOF) break;
        cmd[fp] = (unsigned char)r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp] = 0;

    if (driver->mode != M_ETHERNET && usecrc) {
        unsigned char buf[2];
        driver->getbytes(buf, 2);
        if (driver->done() != DRIVER_SUCCESS) return 1;
        if (buf[1] != 0x00 || r != 0) return frame_error("Open");
        crc_add_block(cmd, fp + 1);
        crc_add_byte(buf[0]);
        if (crc_get()) return crc_error("Open");
    } else {
        if (driver->done() != DRIVER_SUCCESS) return 1;
        if (driver->mode != M_ETHERNET && r != 0) return frame_error("Open");
    }

    if (verbose) {
        gettimeofday(&start, NULL);
    }

    if (debug) {
        log_print("Open:");
        log_hex(cmd);
    }
    buffer_close(buffer);

    path_reset(&outpath);
    if (channel == 15) {
        buffer->mode = CM_ERR;
        commandchannel(cmd);
        status = OPEN_STATUS;
    } else if (cmd[0] == '$') {
        int partitionlist = 0;
        const Petscii *outname;
        const char *path = NULL;
        partition_t partition = 0;
        Partition *outpart;

        buffer->mode = CM_DIR;
        buffer->size = buffer->pointer = 0;
        if (cmd[1] >= '0' && cmd[1] <= '9') {
            int p = 1;
            while (cmd[p] >= '0' && cmd[p] <= '9') partition = partition * 10 + cmd[p++] - '0';
            if (cmd[p] == 0) cmd[1] = 0; else partition = 0;
        } 

        if (cmd[1] == 0) {
            partition_update();
            outpart = partition_get(partition);
            if (outpart == NULL) goto partnf;
            outname = cmd + 1;
            path = path_new(&outpath, outpart->path, 0);
            if (path == NULL) {
                log_print("Open: Out of memory");
                goto partnf;
            }
        } else if (cmd[1] == '=' && cmd[2] == 'P') {
            partition_update();
            path = "=p";
            partitionlist = 1;
            outname = NULL;
        } else {
            outname = resolv_part(cmd + 1, &outpath, &outpart);
            if (outname == NULL) {
            partnf:
                if (verbose) log_printf("Open #%u: Partition not found", channel);
                seterror(ER_DRIVE_NOT_READY, 0); goto vege;
            }
            outname = resolv_path(outname, &outpath, arguments->nameconversion);
            path = outpath.data;
            if (outname == NULL) {
                seterror(ER_PATH_NOT_FOUND, 0); goto vege3;
            }
        }

        if (channel == 1) {
            seterror(ER_NO_CHANNEL, 0); goto vege3;
        } else if (channel > 1) {
            Directory *directory = directory_open(&outpath, arguments->nameconversion, 3);
            if (directory == NULL) {
                directory_error(outpath.data);
                goto vege;
            }
            if (buffer_rawdir(buffer, directory)) {
                errnotochannel15(EIO);
                log_print("Open: Out of memory");
                directory_close(directory);
            } else if (directory_close(directory)) {
                errnotochannel15(errno);
            } else {
                seterror(ER_OK, 0);
            }
            status = OPEN_RONLY;
        } else if (partitionlist) {
            if (buffer_partition(buffer, arguments->default_partition)) {
                errnotochannel15(EIO);
                log_print("Open: Out of memory");
            } else {
                seterror(ER_OK, 0);
            }
            status = OPEN_RONLY;
        } else {
            Directory *directory = directory_open(&outpath, arguments->nameconversion, 2);

            if (directory == NULL) {
                directory_error(outpath.data);
                goto vege;
            }
            if (buffer_cookeddir(buffer, directory, outpart->nr, outname)) {
                errnotochannel15(EIO);
                log_print("Open: Out of memory");
                directory_close(directory);
            } else if (directory_close(directory)) {
                errnotochannel15(errno);
            } else {
                seterror(ER_OK, 0);
            }
            status = OPEN_RONLY;
        }
    vege3:
        if (verbose) log_printf("Open #%u: \"$%s\"", channel, path != NULL ? path : "");
    } else {
        Petscii name[17], type[4] = {'*', 0};
        const Petscii *outname;
        unsigned char overwrite;
        const char *path = NULL;
        int found = 0;
        int b;
        Directory_entry dirent;
        Directory *directory;
        Partition *outpart;

        buffer->mode = CM_CLOSED;
        overwrite = 0;
        for (b = 0; cmd[b]; b++) {
            if (cmd[b] != ':') continue;
            if (cmd[0] == '@') overwrite = 1;
            break;
        }

        if (!cmd[0]) goto vege;

        outname = resolv_part(cmd, &outpath, &outpart);
        if (outname == NULL) {
            if (verbose) log_printf("Open #%u: Partition not found", channel);
            seterror(ER_DRIVE_NOT_READY, 0); goto vege;
        }
        outname = resolv_path(outname, &outpath, arguments->nameconversion);
        path = outpath.data;
        if (outname == NULL) {
            seterror(ER_PATH_NOT_FOUND, 0); goto vege2;
        }
        if (channel == 1) memcpy(type, "PRG", 4);
        convertfilename(outname, ',', name, type, &mode);
        if (channel == 0) mode = 'R';
        if (channel == 1) mode = 'W';

        if (!name[0]) {
            seterror(ER_MISSING_FILENAME, 0); goto vege2;
        }

        directory = directory_open(&outpath, arguments->nameconversion, 0);
        if (directory == NULL) {
            directory_error(path);
            goto vege;
        }

        while (directory_read(directory, &dirent))
        {
            if ((dirent.attrib & (A_ANY | A_CLOSED)) != (A_CLOSED | A_NORMAL)) continue;

            if (!matchname(dirent.name, name)) continue;
            if (!matchname(dirent.filetype, type)) continue;

            path = path_append(&outpath, directory_filename(directory));
            found = 1;
            break;
        }
        directory_close(directory);

        if (!found) {
            Petscii lname[MAX_C64NAME_LEN];
            convertc64name(lname, name, type, arguments->nameconversion);
            path = path_appendc64(&outpath, lname);
            if (mode == 'W' && invalid_c64name(lname)) {
                status = ER_INVALID_FILENAME; goto vege2;
            }
        }

        if (path == NULL) {
            errnotochannel15(EIO);
            log_print("Open: Out of memory");
            goto vege2;
        }

        if (mode == 'W') {
            if (found && !overwrite) {
                seterror(ER_FILE_EXISTS, 0);
            } else {
#if defined _WIN32 || defined __DJGPP__
                int fd = utf8_open(path, O_CREAT | O_WRONLY | (overwrite ? O_TRUNC : O_EXCL) | O_BINARY, S_IRUSR | S_IWUSR);
#else
                int fd = utf8_open(path, O_CREAT | O_WRONLY | (overwrite ? O_TRUNC : O_EXCL), S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif
                if (fd < 0) {
                    errnotochannel15(errno);
                } else {
                    buffers_close_name(path);
                    if (buffer_open(buffer, fd, path, "wb")) {
                        errnotochannel15(errno);
                        close(fd);
                    } else {
                        buffer->mode = CM_COMPAT;
                        status = OPEN_WONLY;
                    }
                }
            }
        } else {
            if (!found) {
                seterror(ER_FILE_NOT_FOUND, 0);
            } else {
                int fd;
                switch (mode) {
                case 'R':
                    fd = utf8_open(path, O_RDONLY | O_BINARY, 0);
                    if (fd >= 0) {
                        buffers_close_name(path);
                        if (!buffer_open(buffer, fd, path, "rb")) {
                            buffer->mode = CM_COMPAT;
                            status = OPEN_RONLY;
                            break;
                        }
                    }
                    errnotochannel15(errno);
                    if (fd >= 0) close(fd);
                    break;
                case 'A':
                    fd = utf8_open(path, O_WRONLY | O_BINARY, 0);
                    if (fd >= 0) {
                        buffers_close_name(path);
                        if (!buffer_open(buffer, fd, path, "ab")) {
                            buffer->mode = CM_COMPAT;
                            status = OPEN_WONLY;
                            break;
                        }
                    }
                    errnotochannel15(errno);
                    if (fd >= 0) close(fd);
                    break;
                default:
                    seterror(ER_FILE_TYPE_MISMATCH, 0);
                }
            }
        }
    vege2:
        if (verbose) log_printf("Open #%u: \"%s\" %c", channel, path != NULL ? path : "", mode);
    }
vege:
    path_free(&outpath);
    if (verbose) {
        log_time("Open: Took", &start);
    }
    driver->sendb(status);
    if (driver->mode == M_ETHERNET) driver->sendb(status);
    return driver->flush() != DRIVER_SUCCESS;
}

int readfile_compat(const Driver *driver, unsigned int flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    unsigned int channel;
    unsigned char status = 0;
    size_t j;
    unsigned int bytes;
    unsigned int l = 0;
    struct timeval start;
    Buffer *buffer;
    unsigned char buf[5];

    driver->getbytes(buf, driver->mode == M_ETHERNET ? 2 : usecrc ? sizeof buf : sizeof buf - 1);
    channel = buf[0] & 0x0f;
    buffer = buffers_get(channel);
    if (driver->done() != DRIVER_SUCCESS) return 1;
    if (driver->mode == M_ETHERNET) {
        bytes = buf[1] != 0 ? buf[1] : 256;
    } else {
        if (buf[usecrc ? sizeof buf - 1 : sizeof buf - 2] != 0x00) return frame_error("Read");
        if (usecrc) {
            crc_add_block(buf, sizeof buf - 1);
            if (crc_get()) return crc_error("Read");
        }
        bytes = buf[1] + buf[2] * 0x100u;
        if (bytes == 0) bytes = 65536;
    }

    if (verbose) log_printf("Read #%u: %d bytes", channel, bytes);
    if (driver->mode != M_ETHERNET) driver->sendb(0);
    if (buffer->mode != CM_DIR && buffer->mode != CM_ERR && (buffer->mode != CM_COMPAT || buffer->file == NULL)) {
        log_print((buffer->mode == CM_CLOSED) ? "Read: Channel not open" : "Read: Not readable");
    error:
        driver->sendb(0);
        driver->sendb(0);
        if (driver->mode != M_ETHERNET) {
            if (usecrc) {
                crc_first(0);
                crc_add_byte(0);
                driver->sendb((unsigned char)crc_get());
            }
            driver->sendb(0);
        }
        return driver->flush() != DRIVER_SUCCESS;
    }
    if (buffer->mode == CM_COMPAT) {

        if (buffer_reserve(buffer, bytes)) {
            log_print("Read: Out of memory");
            goto error;
        }
        bytes = l = buffer_read(buffer, bytes);
        if (feof(buffer->file)) status = 0x42;
        else {
            int c = fgetc(buffer->file);
            if (feof(buffer->file)) status = 0x40;
            ungetc(c, buffer->file);
        }
        j = 0;
    } else {
        size_t length = buffer->size;
        if (buffer->mode == CM_ERR) {
            buffer->data[length++] = '\r';
        }
        j = buffer->pointer;
        if (bytes == (length - j)) status = 0x40;
        else if (bytes > length - j) {
            bytes = length - j; status = 0x42;
        }
        l = bytes;
        buffer->pointer += bytes;
    }

    if (driver->mode == M_ETHERNET) {
        unsigned int i;
        driver->sendb(bytes > 0);
        driver->sendb((unsigned char)bytes);
        for (i = 0; i < bytes; i++) driver->sendb(buffer->data[bytes - i - 1 + j]);
        if (bytes & 1) driver->sendb(status);
        driver->sendb(status);
        driver->sendb(status);
        if (driver->flush() != DRIVER_SUCCESS) return 1;
    } else {
        if (verbose) gettimeofday(&start, NULL);
        driver->sendb((unsigned char)~bytes);
        driver->sendb((unsigned char)(~bytes >> 8));
        driver->sendbytes(buffer->data + j, bytes);
        if (usecrc) {
            crc_first((unsigned char)~bytes);
            crc_add_byte((unsigned char)(~bytes >> 8));
            crc_add_block(buffer->data + j, bytes);
            driver->sendb((unsigned char)crc_get());
        }
        driver->sendb(status);
        if (driver->flush() != DRIVER_SUCCESS) return 1;
        if (verbose) {
            log_speed("Read: Sent", l, &start);
        }
    }
    if (buffer->mode == CM_ERR && buffer->size + 1 >= buffer->pointer) seterror(ER_OK, 0);
    return 0;
}

int closefile_compat(const Driver *driver, int unsigned flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    unsigned int channel;
    Buffer *buffer;
    struct timeval start;
    unsigned char buf[3];

    driver->getbytes(buf, driver->mode == M_ETHERNET ? 1 : usecrc ? sizeof buf : sizeof buf - 1);
    channel = buf[0] & 0x0f;
    buffer = buffers_get(channel);
    if (driver->done() != DRIVER_SUCCESS) return 1;
    if (driver->mode != M_ETHERNET) {
        if (buf[usecrc ? sizeof buf - 1 : sizeof buf - 2] != 0x00) return frame_error("Close");
        if (usecrc) {
            crc_add_block(buf, sizeof buf - 1);
            if (crc_get()) return crc_error("Close");
        }
    }

    if (verbose) {
        gettimeofday(&start, NULL);
    }
    if (verbose) log_printf("Close #%u", channel);
    if (buffer->mode == CM_COMPAT) {
        buffer_close(buffer);
        if (channel == 15) buffer->mode = CM_ERR;
        else {
            buffer->mode = CM_CLOSED;
            buffer_free(buffer);
        }
    }
    if (verbose) {
        log_time("Close: Took", &start);
    }

    driver->sendb(0);
    if (driver->mode == M_ETHERNET) driver->sendb(0);
    return driver->flush() != DRIVER_SUCCESS;
}

int writefile_compat(const Driver *driver, int unsigned flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    unsigned int channel;
    unsigned int bytes;
    struct timeval start, end;
    Buffer *buffer;
    unsigned char buf[5];

    driver->getbytes(buf, driver->mode == M_ETHERNET ? 3 : usecrc ? sizeof buf : sizeof buf - 1);
    channel = buf[0] & 0x0f;
    buffer = buffers_get(channel);
    bytes = buf[1];
    if (driver->mode != M_ETHERNET) {
        if (driver->done() != DRIVER_SUCCESS) return 1;
        if (buf[usecrc ? sizeof buf - 1 : sizeof buf - 2] != 0x00) return frame_error("Write");
        if (usecrc) {
            crc_add_block(buf, sizeof buf - 1);
            if (crc_get()) return crc_error("Write");
        }
        bytes |= buf[2] * 0x100u;
    }

    if (verbose) log_printf("Write #%u: %d bytes", channel, bytes);
    if (buffer->mode == CM_ERR) {
        Petscii cmd[258];

        if (driver->mode == M_ETHERNET) {
            driver->getbytes(cmd, bytes);
            if (driver->done() != DRIVER_SUCCESS) return 1;
            driver->sendb(0);
            driver->sendb(0);
        } else {
            driver->sendb(0);
            if (driver->flush() != DRIVER_SUCCESS) return 1;

            driver->getbytes(cmd, usecrc ? bytes + 2 : bytes + 1);
            if (driver->done() != DRIVER_SUCCESS) return 1;
            if (cmd[usecrc ? bytes + 1 : bytes] != 0x00) return frame_error("Write");
            if (usecrc) {
                crc_clear();
                crc_add_block(cmd, bytes + 1);
                if (crc_get()) return crc_error("Write");
            }

            driver->sendb(0);
        }
        if (driver->flush() != DRIVER_SUCCESS) return 1;
        cmd[bytes - (bytes && cmd[bytes - 1] == 0x0d)] = 0;
        commandchannel(cmd);
        return 0;
    } else {
        if (buffer->mode != CM_COMPAT) {
            driver->done();
            log_print("Write: Not writeable");
        error:
            driver->sendb(2);
            if (driver->mode == M_ETHERNET) driver->sendb(2);
            return driver->flush() != DRIVER_SUCCESS;
        }

        if (buffer_reserve(buffer, driver->mode == M_ETHERNET ? bytes : bytes + 2)) {
            log_print("Write: Out of memory");
            goto error;
        }

        if (driver->mode == M_ETHERNET) {
            driver->getbytes(buffer->data, bytes);
            if (driver->done() != DRIVER_SUCCESS) return 1;
            if (buffer_write(buffer, bytes) == bytes && fflush(buffer->file) == 0) {
                driver->sendb(0);
                driver->sendb(0);
            } else {
                errnotochannel15(errno);
                log_printf("Write: Couldn't write: %s(%d)", strerror(errno), errno);
                driver->sendb(2);
                driver->sendb(2);
            }
            if (driver->flush() != DRIVER_SUCCESS) return 1;
        } else {
            unsigned int totalbytes;
            driver->sendb(0);
            if (driver->flush() != DRIVER_SUCCESS) return 1;

            if (verbose) gettimeofday(&start, NULL);
            totalbytes = usecrc ? bytes + 2 : bytes + 1;
            if (totalbytes >= 1024) {
                unsigned int offs = totalbytes - 512;
                driver->turn(-1);
                driver->getbytes(buffer->data, offs);
                driver->turn(0);
                driver->getbytes(buffer->data + offs, 512);
            } else {
                driver->getbytes(buffer->data, totalbytes);
            }
            if (verbose) gettimeofday(&end, NULL);

            if (driver->done() != DRIVER_SUCCESS) return 1;
            if (buffer->data[totalbytes - 1] != 0x00) return frame_error("Write");
            if (usecrc) {
                crc_clear();
                crc_add_block(buffer->data, bytes + 1);
                if (crc_get()) return crc_error("Write");
            }

            if (buffer_write(buffer, bytes) == bytes && fflush(buffer->file) == 0) {
                driver->sendb(0);
            } else {
                errnotochannel15(errno);
                log_printf("Write: Couldn't write: %s(%d)", strerror(errno), errno);
                driver->sendb(2);
            }
            if (driver->flush() != DRIVER_SUCCESS) return 1;
            if (verbose) {
                log_speed("Write: Received", bytes, &start);
            }
        }
    }
    return 0;
}

