/*

 ideservd.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ideservd.h"
#include <errno.h>
#if defined _WIN32 || defined __CYGWIN__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <commctrl.h>
#include "resource.h"
#else
#include <pwd.h>
#include <grp.h>
#include <sys/wait.h>
#include <signal.h>
#ifndef __DJGPP__
#include <syslog.h>
#endif
#endif
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdlib.h>
#include <locale.h>
#ifdef __linux__
#include <sys/prctl.h>
#endif
#include "path.h"
#include "partition.h"
#include "driver.h"
#include "arguments.h"
#include "crc8.h"
#include "compat.h"
#include "normal.h"
#include "log.h"
#include "message.h"
#include "buffer.h"
#include "directory.h"
#include "petscii.h"
#include "utf8.h"
#include "timeout.h"

static Arguments arguments;

#if defined _WIN32 || defined __CYGWIN__ || defined  __DJGPP__
#else
#define FORKING
static int pipefd = -1;
static pid_t childpid = -1;
#endif

#if defined _WIN32 || defined __CYGWIN__
#else
static void signal_reset2(int signum) {
#if defined _POSIX_C_SOURCE || _POSIX_VERSION >= 199506L
    struct sigaction sa;
    sa.sa_handler = SIG_DFL;
    sa.sa_flags = 0;
    sigemptyset(&sa.sa_mask);
    sigaction(signum, &sa, NULL);
#else
    signal(signum, SIG_DFL);
#endif
}

#if defined _POSIX_C_SOURCE || _POSIX_VERSION >= 199506L
#ifdef SA_RESETHAND
static void signal_reset(int signum) {
    (void)signum;
}
#else
#define signal_reset(s) signal_reset2(s);
#endif
#else
#define signal_reset(s) signal_reset2(s);
#endif

static void signal_set(int signum, void (*handler)(int)) {
#if defined _POSIX_C_SOURCE || _POSIX_VERSION >= 199506L
    struct sigaction sa, osa;
    sa.sa_handler = handler;
    sa.sa_flags = (int)SA_RESETHAND;
    sigemptyset(&sa.sa_mask);
    if (sigaction(signum, NULL, &osa) == 0 && osa.sa_handler != SIG_IGN) {
        sigaction(signum, &sa, NULL);
    }
#else
    if (signal(signum, handler) == SIG_IGN) signal(signum, SIG_IGN);
#endif
}

static const Driver *terminate_driver = NULL;

static void terminate(int x) {
    const Driver *driver;
    signal_reset(x);
    if (!signal_received) {
        signal_received = x;
        signal_set(x, terminate);
#ifdef FORKING
#ifndef PR_SET_PDEATHSIG
        if (childpid > 0) kill(childpid, x);
#endif
#endif
        return;
    }
    driver = terminate_driver;
    terminate_driver = NULL;
    if (driver != NULL) {
        driver->shutdown();
        driver->destroy();
    }
    buffers_free();
    partition_destroy();
#ifdef FORKING
#ifndef PR_SET_PDEATHSIG
    if (childpid > 0) kill(childpid, x);
#endif
    if (pipefd >= 0) close(pipefd);
#endif
    log_close();
    raise(x);
}

static void setup_terminate(const Driver *driver) {
    terminate_driver = driver;
    signal_set(SIGTERM, terminate);
    signal_set(SIGINT, terminate);
#ifdef SIGHUP
    signal_set(SIGHUP, terminate);
#endif
}
#endif

void seterror(Errorcode c, unsigned int i1) {
    Buffer *buffer = buffers_get(15);
    const char *msg;
    int ret;

    switch (c) {
    case ER_OK: msg = "OK"; break;
    case ER_FILES_SCRATCHED: msg = "FILES SCRATCHED"; break;
    case ER_PARTITION_SELECTED: msg = "PARTITION SELECTED"; break;
    case ER_READ_ERROR: if (i1 == 0) msg = "READ ERROR"; else msg = "CRC ERROR"; i1 = 0; break;
    case ER_WRITE_PROTECT_ON: msg = "WRITE PROTECT ON"; break;
    case ER_ACCESS_DENIED: msg = "ACCESS DENIED"; break;
    case ER_UNKNOWN_COMMAND: msg = "UNKNOWN COMMAND"; break;
    case ER_SYNTAX_ERROR:
    case ER_INVALID_FILENAME:
    case ER_MISSING_FILENAME: msg = "SYNTAX ERROR"; break;
    case ER_PATH_NOT_FOUND: msg = "PATH NOT FOUND"; break;
    case ER_FRAME_ERROR: msg = "FRAME ERROR"; break;
    case ER_CRC_ERROR: msg = "CRC ERROR"; break;
    case ER_FILE_NOT_FOUND: msg = "FILE NOT FOUND"; break;
    case ER_FILE_EXISTS: msg = "FILE EXISTS"; break;
    case ER_FILE_TYPE_MISMATCH: msg = "FILE TYPE MISMATCH"; break;
    case ER_NO_CHANNEL: msg = "NO CHANNEL"; break;
    case ER_PARTITION_FULL: msg = "PARTITION FULL"; break;
    case ER_DOS_VERSION: msg = "IDESERV " VERSION2 " PCLINK"; break;
    case ER_DRIVE_NOT_READY: msg = "DRIVE NOT READY"; break;
    case ER_SELECTED_PARTITION_ILLEGAL: msg = "SELECTED PARTITION ILLEGAL"; break;
    default: msg = "UNKNOWN ERROR"; break;
    }
    ret = sprintf((char *)buffer->data, "%02d, %s,%03u,000,000,000", c, msg, i1);
    buffer->size = ret >= 0 ? (unsigned int)ret : 0;
    buffer->pointer = 0;
}

static Errorcode errnotostatus(int i) {
    Errorcode c;
    switch (i) {
    case 0: c = ER_OK; break;
    case EISDIR:
    case ENOTDIR: c = ER_FILE_TYPE_MISMATCH; break;
    case ENOENT: c = ER_FILE_NOT_FOUND; break;
    case EEXIST: c = ER_FILE_EXISTS; break;
    case EPERM:
    case EACCES: c = ER_ACCESS_DENIED; break;
    case ENOTEMPTY: c = ER_FILES_SCRATCHED; break;
    case ENOSPC: c = ER_PARTITION_FULL; break;
    case EROFS: c = ER_WRITE_PROTECT_ON; break;
    case EIO: c = ER_READ_ERROR; break;
    default: c = ER_UNKNOWN_ERROR; break;
    }
    return c;
}

Errorcode errnotochannel15(int i) {
    Errorcode c = errnotostatus(i);
    seterror(c, (c == ER_UNKNOWN_ERROR) ? (unsigned int)i : 0); 
    return c;
}

static Errorcode do_chdir(const Petscii *cmd) {
    Petscii name[17];
    const Petscii *outname;
    Path outpath;
    const char *path = NULL;
    int found = 0;
    int colon;
    Errorcode status = ER_OK;
    Partition *outpart;
    Directory_entry dirent;
    Directory *directory;

    colon = (strchr((char *)cmd, ':') != NULL);
    if (!colon) {
        if (!strcmp((char *)cmd, "_")) {
            cmd = (Petscii *)":_";
            colon = 1;
        } else if (cmd[0] != '/') {
            status = ER_SYNTAX_ERROR; goto vege4;
        } else cmd++;
    }

    outname = resolv_part(cmd, &outpath, &outpart);
    if (outname == NULL) {
        if (arguments.verbose) log_print("Command: Change directory partition not found");
        status = ER_DRIVE_NOT_READY; goto vege4;
    }
    outname = resolv_path(outname, &outpath, arguments.nameconversion);
    path = outpath.data;
    if (outname == NULL) goto vege2;

    convertfilename(outname, ',', name, NULL, NULL);

    if ((colon && name[0] == '_' && name[1] == 0) || (name[0] == '.' && name[1] == '.' && name[2] == 0)) {
        path = path_parent(&outpath);
        found = 1;
        goto vege2;
    }
    if (name[0] == 0 || (name[0] == '.' && name[1] == 0)) {
        found = 1;
        goto vege2;
    }

    directory = directory_open(&outpath, arguments.nameconversion, 0);
    if (directory == NULL) {
        status = errnotostatus(errno);
        goto vege3;
    }

    while (directory_read(directory, &dirent)) {
        if ((dirent.attrib & A_ANY) != A_DIR) continue;

        if (!matchname(dirent.name, name)) continue;

        path = path_append(&outpath, directory_filename(directory));
        found = 1;
        break;
    }
    directory_close(directory);
    if (!found) {
        Petscii lname[MAX_C64NAME_LEN];
        convertc64name(lname, name, NULL, arguments.nameconversion);
        path = path_appendc64(&outpath, lname);
    } else {
        if (directory_check(&outpath)) {
            status = errnotostatus(errno);
            goto vege3;
        }
    }
    if (path == NULL) log_print("Command: Out of memory");
vege2:
    if (found && path != NULL) {
#ifdef FORKING
        if (pipefd >= 0) {
            unsigned char msg[3];
            size_t pathlen = strlen(path);
            if (pathlen < 256) {
                msg[0] = 1;
                msg[1] = (unsigned char)outpart->nr;
                msg[2] = (unsigned char)pathlen;
                write(pipefd, msg, sizeof msg);
                if (msg[2] != 0) write(pipefd, path, msg[2]);
            }
        }
#endif
        partition_set_path(outpart, path);
    } else {
        status = ER_PATH_NOT_FOUND;
    }
vege3:
    if (arguments.verbose) log_printf("Command: Change directory \"%s\"", path == NULL ? "" : path);
    path_free(&outpath);
vege4:
    return status;
}

static Errorcode do_mkdir(const Petscii *cmd) {
    Path outpath;
    const char *path = NULL;
    int found = 0;
    Errorcode status = ER_OK;
    Partition *outpart;
    Petscii name[17];
    const Petscii *outname;
    Directory_entry dirent;
    Directory *directory;

    if (!strchr((char *)cmd, ':')) {
        status = ER_SYNTAX_ERROR; goto vege3;
    }

    outname = resolv_part(cmd, &outpath, &outpart);
    if (outname == NULL) {
        if (arguments.verbose) log_print("Command: Make directory partition not found");
        status = ER_DRIVE_NOT_READY; goto vege3;
    }
    outname = resolv_path(outname, &outpath, arguments.nameconversion);
    path = outpath.data;
    if (outname == NULL) {
        status = ER_PATH_NOT_FOUND; goto vege2;
    }

    convertfilename(outname, ',', name, NULL, NULL);

    if (!name[0]) {
        status = ER_MISSING_FILENAME;
        goto vege2;
    }

    directory = directory_open(&outpath, arguments.nameconversion, 0);
    if (directory == NULL) {
        status = errnotostatus(errno);
        goto vege2;
    }

    while (directory_read(directory, &dirent))
    {
        if ((dirent.attrib & A_ANY) != A_DIR) continue;

        if (!matchname(dirent.name, name)) continue;

        path = path_append(&outpath, directory_filename(directory));
        found = 1;
        break;
    }
    directory_close(directory);

    if (found) {
        status = ER_FILE_EXISTS;
    } else {
        Petscii lname[MAX_C64NAME_LEN];
        convertc64name(lname, name, NULL, arguments.nameconversion);
        path = path_appendc64(&outpath, lname);
        if (invalid_c64name(lname)) {
            status = ER_INVALID_FILENAME; goto vege2;
        }
        if (path == NULL || utf8_mkdir(path)) {
            status = errnotostatus(errno);
        }
    }
    if (path == NULL) log_print("Command: Out of memory");
vege2:
    if (arguments.verbose) log_printf("Command: Make directory \"%s\"", path == NULL ? "" : path);
    path_free(&outpath);
vege3:
    return status;
}

static Errorcode do_scratch(const Petscii *cmd, int smode, int *nr) {
    Path outpath;
    const char *path = NULL;
    unsigned int found = 0;
    Errorcode status = ER_FILES_SCRATCHED;
    Partition *outpart;
    Petscii name[17], type[4] = {'*', 0};
    const Petscii *outname;
    Directory_entry dirent;
    Directory *directory;

    if (!strchr((char *)cmd, ':')) {
        status = ER_SYNTAX_ERROR; goto vege2;
    }

    outname = resolv_part(cmd, &outpath, &outpart);
    if (outname == NULL) {
        if (arguments.verbose) log_print(smode ? "Command: Remove directory partition not found" : "Command: Remove file partition not found");
        status = ER_DRIVE_NOT_READY; goto vege2;
    }
    outname = resolv_path(outname, &outpath, arguments.nameconversion);
    path = outpath.data;
    if (outname == NULL) {
        status = ER_PATH_NOT_FOUND; goto vege;
    }

    convertfilename(outname, '=', name, type, NULL);

    if (!name[0]) {
        status = ER_MISSING_FILENAME;
        goto vege;
    }

    directory = directory_open(&outpath, arguments.nameconversion, 0);
    if (directory == NULL) {
        status = errnotostatus(errno);
        goto vege;
    }

    while (directory_read(directory, &dirent))
    {
        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED)) != (A_DIR | A_CLOSED)) continue;
        } else {
            if ((dirent.attrib & A_ANY) == A_DIR) continue;
        }

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        if (!matchname(dirent.name, name)) continue;
        if (!smode) {
            if (!matchname(dirent.filetype, type)) continue;
        }

        path = directory_path(directory);
        found++;
        if (smode) {
            if (path == NULL || utf8_rmdir(path)) {
                status = errnotostatus(errno);
            }
            if (arguments.verbose) log_printf("Command: Remove directory \"%s\"", path == NULL ? "" : path);
        } else {
            if (path != NULL) buffers_close_name(path);
            if (path == NULL || utf8_unlink(path)) {
                status = errnotostatus(errno);
            }
            if (arguments.verbose) log_printf("Command: Remove file \"%s\"", path == NULL ? "" : path);
        }
    }
    directory_close(directory);
    if (found == 0) {
        Petscii lname[MAX_C64NAME_LEN];
        convertc64name(lname, name, NULL, arguments.nameconversion);
        path = path_appendc64(&outpath, lname);
    }
    if (path == NULL) log_print("Command: Out of memory");
vege:
    if (found == 0) {
        if (smode) {
            if (arguments.verbose) log_printf("Command: Remove directory \"%s\"", path == NULL ? "" : path);
        } else {
            if (arguments.verbose) log_printf("Command: Remove file \"%s\"", path == NULL ? "" : path);
        }
    }
    path_free(&outpath);
vege2:
    *nr = (status == ER_FILES_SCRATCHED) ? found : 0;
    return status;
}

static Errorcode change_partition(partition_t partition) {
    if (arguments.verbose) log_printf("Command: Change to partition %d", partition);
    partition_update();
    if (partition_get(partition) == NULL) {
        return ER_SELECTED_PARTITION_ILLEGAL;
    }
#ifdef FORKING
    if (pipefd >= 0) {
        unsigned char msg[3];
        msg[0] = 2;
        msg[1] = (unsigned char)partition;
        msg[2] = 0;
        write(pipefd, msg, sizeof msg);
    }
#endif
    partition_set_current(partition);
    return ER_PARTITION_SELECTED;
}

void commandchannel(const Petscii *s) {
    Errorcode status = ER_UNKNOWN_COMMAND;
    int nr = 0;
    if (s[0] == 'C' && s[1] == 'D') {
        status = do_chdir(s + 2);
    } else if (s[0] == 'M' && s[1] == 'D') {
        status = do_mkdir(s);
    } else if (s[0] == 'R' && s[1] == 'D') {
        status = do_scratch(s, 1, &nr);
    } else if (s[0] == 'S' && s[1] != '-') {
        status = do_scratch(s, 0, &nr);
    } else if (s[0] == 'C' && s[1] == 0xd0) {
        partition_t part = s[2];
        if (part == 0) part = partition_get_current();
        status = change_partition(part);
        nr = part;
    } else if (s[0] == 'C' && s[1] == 'P') {
        partition_t part = 0;
        int i = 2;
        if (!s[2]) part = partition_get_current(); else {
            while ((s[i] ^ 0x30) < 10) part = part * 10 + (s[i++] ^ 0x30);
        }
        status = change_partition(part);
        nr = part;
    } else if (s[0] == 'U' && ((s[1] & 0x0f) == 9 || (s[1] & 0x0f) == 10)) {
        if (arguments.verbose) log_print("Command: Identify device");
        status = ER_DOS_VERSION;
    } else if (s[0] == 'I') {
        if (arguments.verbose) log_print("Command: Initialize");
        status = ER_OK;
    } else if (s[0] == 'V') {
        if (arguments.verbose) log_print("Command: Validate");
        status = ER_OK;
    } else if (s[0] == 'M' && s[1] == '-' && s[2] == 'R') {
        Buffer *buffer = buffers_get(15);
        int i, j, k;
        j = s[5]; if (j > 50) j = 50;
        k = s[3];
        for (i = 0; i < j; i++) buffer->data[i] = (unsigned char)"IDE64 CARTRIDGE "[k++ & 15];
        buffer->size = i;
        buffer->pointer = 0;
        if (arguments.verbose) log_printf("Command: Memory read $%04x %02x", s[3] | (s[4] << 8), s[5]);
        return;
    } else if (s[0] == 0) {
        return;
    }
    seterror(status, nr);
}

typedef struct Environment {
    const char *root;
#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
#else
    int setuid;
    uid_t uid;
    int setgid;
    gid_t gid;
#endif
} Environment;

static int prepare_environment(Environment *environment) {
    environment->root = arguments.root;
    if (arguments.root != NULL) {
        if (chdir(arguments.root) != 0) {
            message("Changing dir to \"%s\" failed: %s(%d)\n", arguments.root, strerror(errno), errno);
            return 1;
        }
    }
#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
#ifdef _WIN32
    unc_path = (GetVersion() & 0x80000000) == 0;
#endif
#else 
    if (arguments.root != NULL) {
        time_t now = time(NULL);
        localtime(&now);
    }
    errno = 0;
    if (arguments.priority && nice(arguments.priority) == -1 && errno != 0) {
        message("Priority adjustment failed: %s(%d)\n", strerror(errno), errno);
        return 1;
    }
#ifndef __HAIKU__
    if (arguments.user == NULL || arguments.group == NULL) {
        uid_t uid = getuid(), euid = geteuid();
        if (uid <= 0 || uid != euid) {
            if (arguments.user == NULL) arguments.user = "nobody";
            if (arguments.group == NULL) arguments.group = "nogroup";
        }
    }
#endif
    environment->setuid = arguments.user != NULL;
    if (environment->setuid) {
        struct passwd *user2;
        errno = 0;
        user2 = getpwnam(arguments.user);
        if (user2 == NULL) {
            if (errno == 0 || errno == ENOENT || errno == ESRCH || errno == EBADF || errno == EPERM) {
                message("No such user name: %s\n", arguments.user);
            } else {
                message("User name search failed: %s(%d)\n", strerror(errno), errno);
            }
            return 1;
        }
        environment->uid = user2->pw_uid;
    }
    environment->setgid = arguments.group != NULL;
    if (environment->setgid) {
        struct group *group2;
        errno = 0;
        group2 = getgrnam(arguments.group);
        if (group2 == NULL) {
            if (errno == 0 || errno == ENOENT || errno == ESRCH || errno == EBADF || errno == EPERM) {
                message("No such group name: %s\n", arguments.group);
            } else {
                message("Group name search failed: %s(%d)\n", strerror(errno), errno);
            }
            return 1;
        }
        environment->gid = group2->gr_gid;
    }
#endif
    return 0;
}

static int setup_environment(Environment *environment) {
#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
    (void)environment;
#else
    if (environment->root != NULL) {
        uid_t uid = getuid(), euid = geteuid();
        if (chroot(environment->root) != 0 || chdir("/") != 0) {
            log_printf("Chrooting failed: %s(%d)", strerror(errno), errno);
            if (uid <= 0 || uid != euid) {
                return 1;
            }
        }
    }
    if (environment->setgid) {
        gid_t gid;
        int groups = getgroups(0, NULL);
        if (groups != 0 && (groups != 1 || getgroups(1, &gid) == -1 || environment->gid != gid)) {
            if (setgroups(1, &environment->gid) != 0) {
                log_print("Could not drop all groups");
                return 1;
            }
        }
        if (setgid(environment->gid) != 0 || environment->gid != getgid() || environment->gid != getegid()) {
            log_printf("Could not change group: %s(%d)", strerror(errno), errno);
            return 1;
        }
        groups = getgroups(0, NULL);
        if (groups != 0 && (groups != 1 || getgroups(1, &gid) == -1 || environment->gid != gid)) {
            log_print("Could not drop all groups");
            return 1;
        }
    }
    if (environment->setuid) {
        if (setuid(environment->uid) != 0 || environment->uid != getuid() || environment->uid != geteuid()) {
            log_printf("Could not change user: %s(%d)", strerror(errno), errno);
            return 1;
        }
    }
#endif
    if (buffers_init()) {
        log_print("Out of memory");
        return 1;
    }
    seterror(ER_DOS_VERSION, 0);
    return 0;
}

static int rate_limit(struct timeval *start) {
    struct timeval end;
    if (gettimeofday(&end, NULL) == -1) return 1;
    if (end.tv_sec < start->tv_sec) return 1;
    if (start->tv_sec == end.tv_sec) {
        if (end.tv_usec < start->tv_usec) return 1;
        end.tv_usec -= start->tv_usec;
        if (end.tv_usec < 1000000) udelay(1000000 - end.tv_usec);
    } else if (start->tv_sec + 1 == end.tv_sec) {
        if (start->tv_usec > end.tv_usec) {
            end.tv_usec = start->tv_usec - end.tv_usec;
            if (end.tv_usec < 1000000) udelay(end.tv_usec);
        }
    }
    return 0;
}

static void partition_init(void) {
#if defined _WIN32 || defined __CYGWIN__
    if (arguments.root == NULL) {
        partition_t partition = 0, current;
        DWORD drives = GetLogicalDrives() & 0x3ffffff;

        for (current = 1; drives != 0; current++, drives >>= 1) {
            if ((drives & 1) == 0) continue;
            if (arguments.default_partition == (int)partition && partition != 0) continue;
            if (partition < 3 || arguments.default_partition == (int)current) {
                partition = current;
            }
        }
        if (partition == 0) {
            partition = 1;
            partition_create(partition, (Petscii *)"PARTITION 1", "");
        } else {
            partition_set_auto();
        }
        arguments.default_partition = (int)partition;
        partition_set_current(partition);
        return;
    } else {
#ifdef _WIN32
        char *path = utf8_getcwd();
#else
        char *path = getcwd(NULL, 0);
#endif
        partition_create(1, (Petscii *)"PARTITION 1", path != NULL ? path : "/");
        free(path);
    }
#else
    partition_create(1, (Petscii *)"PARTITION 1", "");
#endif
    arguments.default_partition = 1;
    partition_set_current(1);
}

static void driver_initialize(const Driver *driver) {
    log_flush();
    do {
        struct timeval start;
        int ret = gettimeofday(&start, NULL);
        int lastfail = driver->initialize();
        if (lastfail == 0) {
            return;
        }
        log_flush();
        if (signal_received) break;
        if (ret == -1 || rate_limit(&start)) {
            udelay(1000000);
        }
    } while (!signal_received);
}

static void serve(const Driver *driver) {
    int b;
    unsigned char ec = 0x5a;
    unsigned int flags = 0;
    if (arguments.verbose) flags |= FLAG_VERBOSE;
    if (arguments.debug) flags |= FLAG_DEBUG;

    while (!signal_received) {
        log_flush();
    start2:
        b = driver->wait(ec);
        if (b < 0) return;

        switch (b) {
        case 0x00: goto start2;
        case 0xCE: crc_first(b); /* fall through */
        case 'N': b = openfile(driver, &arguments, (b == 0xCE ? FLAG_USE_CRC : 0) | flags); ec = 0x5a; break;
        case 0xC7: crc_first(b);  /* fall through */
        case 'G': b = readfile(driver, (b == 0xC7 ? FLAG_USE_CRC : 0) | flags); ec = 0x5a; break;
        case 0xD3: crc_first(b); /* fall through */
        case 'P':
        case 'S': b = writefile(driver, (b == 0xD3 ? FLAG_USE_CRC : b == 'P' ? FLAG_USE_PADDING : 0) | flags); ec = 0x5a; break;
        case 0xC4: crc_first(b); /* fall through */
        case 'D': b = closefile(driver, (b == 0xC4 ? FLAG_USE_CRC : 0) | flags); ec = 0x5a; break;
        case 0xC9: crc_first(b); /* fall through */
        case 'I': b = statuserror(driver, (b == 0xC9 ? FLAG_USE_CRC : 0) | flags); ec = 0x5a; break;
        case 0xCF: crc_first(b); /* fall through */
        case 'O': b = openfile_compat(driver, &arguments, (b == 0xCF ? FLAG_USE_CRC : 0) | flags); ec = 0; break;
        case 0xD2: crc_first(b); /* fall through */
        case 'R': b = readfile_compat(driver, (b == 0xD2 ? FLAG_USE_CRC : 0) | flags); ec = 0; break;
        case 0xD7: crc_first(b); /* fall through */
        case 'W': b = writefile_compat(driver, (b == 0xD7 ? FLAG_USE_CRC : 0) | flags); ec = 0; break;
        case 0xC3: crc_first(b); /* fall through */
        case 'C': b = closefile_compat(driver, (b == 0xC3 ? FLAG_USE_CRC : 0) | flags); ec = 0; break;
        default: log_printf("Unknown command %02X", b);
        }
        if (b != 0) {
            if (driver->clean()) log_print("Timeout");
        }
    }
}

#if defined _WIN32 || defined __CYGWIN__
static void nologwindow_append(void *param, const char *text) {
    (void)param;
    fputs(text, stderr);
    fputc('\n', stderr);
}

static void logwindow_append(void *param, const char *text) {
    HWND hwnd = *(HWND *)param;
#ifdef _WIN32
    if (!IsWindow(hwnd)) return;
    if (use_wide) {
        wchar_t *data = utf8_to_wchar(text);
        if (data != NULL) SendMessageW(hwnd, WM_USER + 1, 0, (LPARAM)(LPVOID)data);
    } else {
        char *data = utf8_to_char(text);
        if (data != NULL) SendMessage(hwnd, WM_USER + 1, 0, (LPARAM)(LPVOID)data);
    }
#else
    char *data;
    size_t len;
    if (!IsWindow(hwnd)) return;
    len = strlen(text) + 1;
    data = (char *)malloc(len);
    if (data == NULL) return;
    memcpy(data, text, len);
    SendMessage(hwnd, WM_USER + 1, 0, (LPARAM)(LPVOID)data);
#endif
}

static BOOL CALLBACK MyDlgProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    static UINT_PTR timer;
    static HWND logwindow_handle;
    if (logwindow_handle == NULL) logwindow_handle = GetDlgItem(hwnd, IDC_LOG);
    switch (msg) {
    case WM_SYSCOMMAND:
        switch (wParam) {
        case IDC_HIDE:
            ShowWindow(hwnd, SW_HIDE);
            return TRUE;
        }
        break;
    case WM_COMMAND:
        switch (wParam) {
        case IDOK:
        case IDCANCEL:
        case IDC_EXIT:
            if (timer != 0) {
                KillTimer(hwnd, timer);
                timer = 0;
            }
            PostQuitMessage(0);
            return TRUE;
        case IDC_HIDE:
            ShowWindow(hwnd, SW_HIDE);
            return TRUE;
        }
        break;
    case WM_USER + 1:
        {
            int len = GetWindowTextLength(logwindow_handle);
            static int max_size;
            if (max_size == 0) max_size = SendMessage(logwindow_handle, EM_GETLIMITTEXT, 0, 0);
            else if (timer == 0) SendMessage(logwindow_handle, WM_SETREDRAW, FALSE, 0);
            while (len + 256 > max_size && max_size > 512) {
                SendMessage(logwindow_handle, EM_SETSEL, 0, SendMessage(logwindow_handle, EM_LINELENGTH, 0, 0) + 2);
                SendMessage(logwindow_handle, EM_REPLACESEL, 0, (LPARAM)"");
                len = GetWindowTextLength(logwindow_handle);
            }
            if (len != 0) {
                SendMessage(logwindow_handle, EM_SETSEL, len, len);
                SendMessage(logwindow_handle, EM_REPLACESEL, 0, (LPARAM)"\r\n");
                len += 2;
            }
#ifdef _WIN32
            if (use_wide) SendMessageW(logwindow_handle, EM_REPLACESEL, 0, lParam);
            else SendMessage(logwindow_handle, EM_REPLACESEL, 0, lParam);
#else
            SendMessage(logwindow_handle, EM_REPLACESEL, 0, lParam);
#endif
            SendMessage(logwindow_handle, EM_SETSEL, len, len);
            if (timer == 0) timer = SetTimer(hwnd, 1, 100, NULL);
            free((void *)lParam);
        }
        return TRUE;
    case WM_PAINT:
        if (timer != 0) {
            SendMessage(logwindow_handle, WM_SETREDRAW, TRUE, 0);
            KillTimer(hwnd, timer);
            timer = 0;
        }
        break;
    case WM_TIMER:
        if (timer != 0) {
            SendMessage(logwindow_handle, WM_SETREDRAW, TRUE, 0);
            SendMessage(logwindow_handle, EM_SCROLLCARET, 0, 0);
            KillTimer(hwnd, timer);
            timer = 0;
        }
        break;
    case WM_SETCURSOR:
        if (LOWORD(lParam) == HTCLIENT && timer != 0) return TRUE;
        break;
    case WM_SIZE:
        {
            RECT r;
            int bottom_margin, top_margin;
            int width = LOWORD(lParam);
            int height = HIWORD(lParam);
            if (timer != 0) {
                SendMessage(logwindow_handle, WM_SETREDRAW, TRUE, 0);
                KillTimer(hwnd, timer);
                timer = 0;
            }
            r.left = 0; r.top = 0; r.right = 170; r.bottom = 105;
            MapDialogRect(hwnd, &r);
            bottom_margin = r.bottom;
            top_margin = r.bottom;
            r.left = 0; r.top = 0; r.right = 170; r.bottom = 80;
            MapDialogRect(hwnd, &r);
            bottom_margin -= r.bottom;
            MoveWindow(logwindow_handle, r.left, r.top, width, height - bottom_margin, TRUE);

            r.left = 5; r.top = 85; r.right = 5 + 95; r.bottom = 85 + 15;
            MapDialogRect(hwnd, &r);
            top_margin -= r.top;
            MoveWindow(GetDlgItem(hwnd, IDC_HIDE), r.left, height - top_margin, r.right - r.left, r.bottom - r.top, TRUE);
            r.left = 105; r.top = 85; r.right = 105 + 60; r.bottom = 85 + 15;
            MapDialogRect(hwnd, &r);
            MoveWindow(GetDlgItem(hwnd, IDC_EXIT), r.left, height - top_margin, r.right - r.left, r.bottom - r.top, TRUE);
        }
        break;
    case WM_USER:
        if (lParam == WM_RBUTTONDOWN || lParam == WM_LBUTTONDOWN) {
            SetWindowPos(hwnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
            SetWindowPos(hwnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_SHOWWINDOW | SWP_NOSIZE | SWP_NOMOVE);
            SetForegroundWindow(hwnd);
            SetFocus(hwnd);
            SetActiveWindow(hwnd);
            ShowWindow(hwnd, SW_RESTORE);
            SendMessage(logwindow_handle, EM_SCROLLCARET, 0, 0);
            return TRUE;
        }
        break;
    }
    return FALSE;
}

static VOID WINAPI PCLinkTerminate(ULONG_PTR dwParam)
{
    (void)dwParam;
    signal_received = 1;
}

static DWORD WINAPI PCLinkLoop(LPVOID lpParam) {
    const Driver *driver = (const Driver *)lpParam;
    partition_init();
    do {
        driver_initialize(driver);
        if (!signal_received) serve(driver);
        driver->shutdown();
    } while (!signal_received);
    buffers_free();
    partition_destroy();
    return 0;
}
#endif

#ifdef LOG_DAEMON
static void log_syslog(void *param, const char *text) {
    (void)param;
    syslog(LOG_INFO, "%s", text);
}
#endif

int main(int argc, char *argv[]) {
    const Driver *driver;
    Environment environment;
#if defined _WIN32 || defined __CYGWIN__
    HWND hwnd;
    DWORD iID;
#ifdef HAVE_COMCTL
    InitCommonControls();
#endif
#endif

    setlocale(LC_CTYPE, "");

#if defined _WIN32 || defined __CYGWIN__
    hwnd = CreateDialogW(NULL, MAKEINTRESOURCEW(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
#ifdef _WIN32
    use_wide = (hwnd != NULL);
    if (!use_wide) hwnd = CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
#endif
    iID = GetLastError();
    log_callback(Log_callback(hwnd == NULL ? nologwindow_append : logwindow_append, &hwnd));

    log_print("IDEDOS 0.9x PCLink ideservd " VERSION "\r\n"
    "Written by Kajtar Zsolt <soci@c64.rulez.org>\r\n"
    "Win32 port by J.Soucek <josef@ide64.org>\r\n"
    "http://idedos.ide64.org/\r\n"
    "This program is free software. See the GNU\r\n"
    "General Public License for more details.\r\n");
#endif

    testarg(&arguments, argc, argv);
#ifdef LOG_DAEMON
    if (arguments.syslog) {
        openlog("ideservd", 0, LOG_DAEMON);
        log_callback(Log_callback(log_syslog, NULL));
    }
#endif
    log_open(arguments.log);
    log_print("ideservd " VERSION);

    driver = get_driver(&arguments);
    if (driver == NULL) {
        message("No such driver: %s\n", arguments.driver);
        return EXIT_FAILURE;
    }

    if (prepare_environment(&environment)) return EXIT_FAILURE;

#if defined _WIN32 || defined __CYGWIN__
    if (setup_environment(&environment)) return EXIT_FAILURE;
    if (hwnd == NULL) {
        log_printf("CreateDialog failed: %s(%u)", errormsg(iID), (unsigned int)iID);
        PCLinkLoop((LPVOID)driver);
    } else {
        NOTIFYICONDATA notifyIconData;
        BOOL icon;
        HANDLE hThread;
        HMENU hMenu;
        if (arguments.background != 1) ShowWindow(hwnd, SW_SHOWNORMAL);
        memset(&notifyIconData, 0, sizeof notifyIconData);
        notifyIconData.hIcon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICON));
        if (notifyIconData.hIcon == NULL) {
            icon = FALSE;
        } else {
            SendMessage(hwnd, WM_SETICON, ICON_BIG, (LPARAM)notifyIconData.hIcon);
            notifyIconData.cbSize = sizeof notifyIconData;
            notifyIconData.hWnd = hwnd;
            notifyIconData.uID = 0;
            notifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
            notifyIconData.uCallbackMessage = WM_USER;
            strcpy(notifyIconData.szTip, "IDE64 PCLink");
            icon = Shell_NotifyIcon(NIM_ADD, &notifyIconData);
        }
        hMenu = GetSystemMenu(hwnd, FALSE);
        if (hMenu != NULL) {
            InsertMenu(hMenu, 5, MF_BYPOSITION | MF_SEPARATOR, 0, NULL);
            InsertMenu(hMenu, 6, MF_BYPOSITION, IDC_HIDE, TEXT("Hide"));
        }

        hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)PCLinkLoop, (LPVOID)driver, 0, &iID);
        if (hThread != NULL) {
            MSG msg;
            while (GetMessage(&msg, NULL, 0, 0) != 0) {
                if (!IsDialogMessage(hwnd, &msg)) {
                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }
            }
            if (QueueUserAPC(PCLinkTerminate, hThread, 0) == 0) {
                TerminateThread(hThread, 0);
            } else {
                ShowWindow(hwnd, SW_HIDE);
                if (WaitForSingleObject(hThread, 2000) == WAIT_TIMEOUT) {
                    TerminateThread(hThread, 0);
                }
            }
            CloseHandle(hThread);
        }
        if (icon) {
            Shell_NotifyIcon(NIM_DELETE, &notifyIconData);
        }
        log_callback(Log_callback(NULL, NULL));
        DestroyWindow(hwnd);
    }
#elif defined FORKING
    if (arguments.background == 1) {
        pid_t pid = fork();
        if (pid == -1) {
            message("Could not fork to background: %s(%d)\n", strerror(errno), errno);
            return EXIT_FAILURE;
        }
        if (pid != 0) {
            message("Forking into background\n");
            return EXIT_SUCCESS;
        }
        if (setsid() == -1) {
            message("Setsid failed: %s(%d)\n", strerror(errno), errno);
            return EXIT_FAILURE;
        }
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
    }
    partition_init();
    setup_terminate(driver);
    if (arguments.background >= 0) {
#ifdef PR_SET_PDEATHSIG
        pid_t ppid = getpid();
#endif
        do {
            struct timeval start;
            int ret;
            pid_t pid;
            int status;
            int pipefds[2];

            if (pipe(pipefds) == -1) {
                pipefds[0] = -1;
                pipefds[1] = -1;
            }
            pid = fork();
            if (pid == -1) {
                log_print("Fork failed");
                return EXIT_FAILURE;
            }
            if (pid == 0) {
#ifdef PR_SET_PDEATHSIG
                prctl(PR_SET_PDEATHSIG, SIGHUP);
                if (getppid() != ppid) return EXIT_FAILURE;
#endif
                if (pipefds[0] >= 0) close(pipefds[0]);
                pipefd = pipefds[1];
                driver_initialize(driver);
                if (!signal_received && !setup_environment(&environment)) serve(driver);
                terminate_driver = NULL;
                driver->shutdown();
                buffers_free();
                partition_destroy(); /* copy */
                if (pipefds[1] >= 0) close(pipefds[1]);
                if (signal_received) {
                    log_close();
                    signal_reset2(signal_received);
                    raise(signal_received);
                }
                return EXIT_SUCCESS;
            }
            childpid = pid;
            ret = gettimeofday(&start, NULL);
            if (pipefds[1] >= 0) close(pipefds[1]);
            if (pipefds[0] >= 0) {
                unsigned char msg[3];
                while (read(pipefds[0], &msg, sizeof msg) == sizeof msg) {
                    if (msg[0] == 1) {
                        Partition *p;
                        char path[256];
                        ssize_t c = (msg[2] != 0) ? read(pipefds[0], path, msg[2]) : 0;
                        if (c != msg[2]) break;
                        path[c] = 0;
                        partition_update();
                        p = partition_get(msg[1]);
                        if (p != NULL) partition_set_path(p, path);
                    } else if (msg[0] == 2) {
                        partition_set_current(msg[1]);
                    } else {
                        break;
                    }
                }
                close(pipefds[0]);
            }
            while (!signal_received && waitpid(pid, &status, 0) == -1) {
                if (errno == EINTR) continue;
                log_print("Waitpid failed");
                return EXIT_FAILURE;
            }
            childpid = -1;
            if (WIFSIGNALED(status)) {
                int sig = WTERMSIG(status);
                if (!signal_received) {
                    log_printf("Child caught signal %d", sig);
                    signal_received = sig;
                }
            }
            if (signal_received) break;
            if (ret == -1 || rate_limit(&start)) {
                udelay(1000000);
            }
        } while (!signal_received);
    } else {
        driver_initialize(driver);
        if (!signal_received) {
            if (setup_environment(&environment)) return EXIT_FAILURE;
            serve(driver);
        }
        terminate_driver = NULL;
        driver->shutdown();
        buffers_free();
    }
    partition_destroy();
#else
#ifndef __DJGPP__
    if (arguments.background == 1) {
        message("Not possible to run in background\n");
        return EXIT_FAILURE;
    }
#endif
    partition_init();
    setup_terminate(driver);
    if (setup_environment(&environment)) return EXIT_FAILURE;
    do {
        driver_initialize(driver);
        if (!signal_received) serve(driver);
        terminate_driver = NULL;
        driver->shutdown();
    } while (!signal_received);
    buffers_free();
    partition_destroy();
#endif 
    driver->destroy();
    return EXIT_SUCCESS;
}
