/*

 log.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "log.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "utf8.h"

static FILE *lf;
static Log_callback callback;

static char *last_line;
static size_t last_len;
static char *this_line;
static size_t this_len;

static int duplicate(const char *text) {
    size_t len;
    if (last_line != NULL && strcmp(text, last_line) == 0) return 1;
    len = strlen(text) + 1;
    if (last_line == NULL) {
        last_line = (char *)malloc(len);
        if (last_line == NULL) return 0;
        last_len = len;
    } else if (len > last_len) {
        char *line = (char *)realloc(last_line, len);
        if (line == NULL) {
            last_line[0] = 0;
            return 0;
        }
        last_line = line;
        last_len = len;
    }
    memcpy(last_line, text, len);
    return 0;
}

static void write_line(const char *text) {
    if (callback.function != NULL) {
        callback.function(callback.parameter, text);
    }
    if (lf != NULL) {
#ifdef _WIN32
        char *txt2 = utf8_to_char(text);
        if (txt2 == NULL) return;
        fputs(txt2, lf);
        putc('\n', lf);
        free(txt2);
#else
        fputs(text, lf);
        putc('\n', lf);
#endif
    }
}

void log_close(void) {
    if (lf != NULL) {
        fclose(lf);
        lf = NULL;
    }
    if (last_line != NULL) {
        free(last_line);
        last_line = NULL;
    }
    if (this_line != NULL) {
        free(this_line);
        this_line = NULL;
    }
}

void log_callback(Log_callback cb) {
    callback = cb;
}

void log_open(const char *filename) {
    atexit(log_close);

    if (lf != NULL) {
        fclose(lf);
    }
    if (filename == NULL) {
        lf = NULL;
        return;
    }
    if (filename[0] == '-' && filename[1] == 0) {
        lf = stdout;
        setvbuf(lf, NULL, _IOLBF, 256);
        return;
    }
    lf = fopen(filename, "wt");
    if (lf == NULL) {
        if (callback.function == NULL) {
            lf = stdout;
        }
        log_printf("Cannot open logfile \"%s\"", filename);
        lf = NULL;
    }
}

void log_print(const char *text) {
    if (duplicate(text)) return;
    write_line(text);
}

void log_printf(const char *format, ...) {
    int ret;
    va_list args;
    va_start(args, format);
    ret = vsnprintf(this_line, this_len, format, args);
    va_end(args);
    if (ret >= 0 && (unsigned int)ret >= this_len) {
        char *line = (char *)realloc(this_line, (size_t)ret + 1);
        if (line == NULL) {
            return;
        }
        this_line = line;
        this_len = (size_t)ret + 1;
        va_start(args, format);
        vsnprintf(this_line, this_len, format, args);
        va_end(args);
    }
    log_print(this_line);
}

static unsigned long long duration(const struct timeval *start) {
    struct timeval end;
    if (gettimeofday(&end, NULL) == -1) return 0;
    if (end.tv_sec < start->tv_sec) return 0;
    end.tv_sec -= start->tv_sec;
    if (end.tv_sec == 0) {
        if (end.tv_usec < start->tv_usec) return 0;
        return (unsigned long long)(end.tv_usec - start->tv_usec);
    }
    return (unsigned long long)end.tv_sec * 1000000ull + (unsigned long long)end.tv_usec - (unsigned long long)start->tv_usec;
}

void log_time(const char *name, const struct timeval *start) {
    unsigned long long l2 = duration(start);
    const char *unit;
    int p;
    unsigned int l;
    if (l2 >= 1000000) {
        l2 = (l2 + 500) / 1000;
        unit = "";
    } else if (l2 >= 1000) {
        unit = "milli";
    } else if (l2 != 0) {
        unit = "micro";
    } else {
        duplicate("");
        return;
    }
    if (l2 >= 1000) {
        l = (unsigned int)(l2 % 1000);
        if (l == 0) {
            p = 0;
            l = (unsigned int)(l2 / 1000);
        } else if (l % 100 == 0) {
            p = 1;
            l /= 100;
        } else if (l % 10 == 0) {
            p = 2;
            l /= 10;
        } else {
            p = 3;
        }
    } else {
        l = (unsigned int)l2;
        p = 0;
    }
    if (p == 0) {
        log_printf("%s %u %sseconds", name, l, unit);
    } else {
        log_printf("%s %u.%0*u %sseconds", name, (unsigned int)(l2 / 1000), p, l, unit);
    }
}

void log_speed(const char *name, unsigned int bytes, const struct timeval *start) {
    unsigned long long l2 = duration(start);
    const char *unit;
    int p;
    unsigned int l;
    if (l2 != 0) {
        l2 = bytes * 1000000ull / l2;
    }

    if (l2 >= 1000000) {
        l2 = (l2 + 500) / 1000;
        unit = "mega";
    } else if (l2 >= 1000) {
        unit = "kilo";
    } else if (l2 != 0) {
        unit = "";
    } else {
        log_printf("%s %u bytes", name, bytes);
        return;
    }
    if (l2 >= 1000) {
        l = (unsigned int)(l2 % 1000);
        if (l == 0) {
            p = 0;
            l = (unsigned int)(l2 / 1000);
        } else if (l % 100 == 0) {
            p = 1;
            l /= 100;
        } else if (l % 10 == 0) {
            p = 2;
            l /= 10;
        } else {
            p = 3;
        }
    } else {
        l = (unsigned int)l2;
        p = 0;
    }
    if (p == 0) {
        log_printf("%s %u bytes at %u %sbytes/seconds", name, bytes, l, unit);
    } else {
        log_printf("%s %u bytes at %u.%0*u %sbytes/seconds", name, bytes, (unsigned int)(l2 / 1000), p, l, unit);
    }
}

static void writehex(char *msg, unsigned char c) {
    const char *hex = "0123456789ABCDEF";
    msg[0] = hex[c >> 4];
    msg[1] = hex[c & 15];
    msg[2] = ' ';
}

void log_hex(const unsigned char *cmd) {
    char msg[16 * 3];
    int i = 0;
    while (cmd[i]) {
        writehex(msg + (i & 15) * 3, cmd[i]);
        i++;
        if (!(i & 15)) {
            msg[16 * 3 - 1] = '\0';
            write_line(msg);
        }
    }
    if (i & 15) {
        msg[(i & 15) * 3 - 1] = '\0';
        write_line(msg);
    }
    duplicate("");
}

void log_flush(void) {
    if (lf != NULL) {
        fflush(lf);
    }
}
