/*

 usb.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "usb.h"
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "log.h"
#include "driver.h"
#include "timeout.h"

#ifdef HAVE_USB
#if defined _WIN32 || defined __CYGWIN__
#include <stdlib.h>
#include "ftd2xxl.h"
#else
#include <ftdi.h>
#endif
#if defined HAVE_LIBUSB && defined __linux__
#include <libusb.h>
#if defined(LIBUSB_API_VERSION) && (LIBUSB_API_VERSION >= 0x01000102)
#define LIBUSB_HOTPLUG
#endif
#endif

#ifndef FTD2XX_API
static struct ftdi_context *ftDevice;
#if defined LIBFTDI_VERSION
#elif defined LIBFTDI_VERSION0_20
#else
#define LIBFTDI_LATENCY
static int last_latency;
#endif
#else
FT_HANDLE ftHandle;
static int ftd2xx_loaded;
#endif
#ifdef LIBUSB_HOTPLUG
static struct libusb_context *usb;
static libusb_hotplug_callback_handle hotplug_handle;
#endif

static const char *i_dev;
static int inited;
static Driver_errno driver_errno = DRIVER_EIO;
static unsigned char ebufo[4096];
static unsigned int ebufop;
static int last_ec;

static Driver_errno flush(void);

#if defined LIBFTDI_VERSION
#elif defined LIBFTDI_VERSION0_20
#elif defined LIBFTDI_VERSION1_0
#elif defined LIBFTDI_VERSION1_1
#elif defined LIBFTDI_VERSION1_2
#elif defined LIBFTDI_VERSION1_3
#elif defined LIBFTDI_VERSION1_4
#else
#define ftdi_usb_purge_buffers ftdi_tcioflush
#endif

#ifdef LIBUSB_HOTPLUG
static int LIBUSB_CALL hotplug(libusb_context *ctx, libusb_device *device, libusb_hotplug_event event, void *user_data) {
    (void)ctx;
    (void)device;
    (void)event;
    int *hotplug_found = (int *)user_data;
    *hotplug_found = 1;
    return 1;
}
#endif

static int initialize(void) {
#ifndef FTD2XX_API
    struct ftdi_device_list *list, *i, *ad, *d;
    char desc[256], serial[256];
    char desc2[256], serial2[256];

    driver_errno = DRIVER_EIO;
    inited = 0;

    if (ftDevice == NULL) {
        ftDevice = ftdi_new();
    }
    if (ftDevice == NULL) {
        log_print("Out of memory");
        return 1;
    }

#ifdef LIBUSB_HOTPLUG
    if (usb == NULL) {
        int error = libusb_init(&usb);
        if (error != LIBUSB_SUCCESS) {
            log_printf("libusb_init: %s", libusb_strerror(error));
            return 1;
        }
    }
    if (libusb_has_capability(LIBUSB_CAP_HAS_HOTPLUG)) {
        int hotplug_found = 0;
        if (libusb_hotplug_register_callback(usb, LIBUSB_HOTPLUG_EVENT_DEVICE_ARRIVED, LIBUSB_HOTPLUG_ENUMERATE, 0x403, 0x6001, LIBUSB_HOTPLUG_MATCH_ANY, hotplug, &hotplug_found, &hotplug_handle) != LIBUSB_SUCCESS) {
            log_print("Failed to register hotplug callback");
        } else if (hotplug_found) {
            libusb_hotplug_deregister_callback(usb, hotplug_handle);
        } else {
            log_print("No usable USB device found for now");
            do {
                if (libusb_handle_events_completed(usb, NULL) != LIBUSB_SUCCESS) {
                    if (!hotplug_found) libusb_hotplug_deregister_callback(usb, hotplug_handle);
                    return 1;
                }
            } while (!hotplug_found);
        }
    }
#endif
    if (ftdi_usb_find_all(ftDevice, &list, 0x0403, 0x6001) < 0) {
        log_printf("ftdi_usb_find_all: %s", ftdi_get_error_string(ftDevice));
        return 1;
    }

    ad = NULL; d = NULL;
    for (i = list; i != NULL; i = i->next) {
        if (ftdi_usb_get_strings(ftDevice, i->dev, NULL, 0, desc, sizeof desc, serial, sizeof serial) < 0) {
            continue;
        }
        if (d == NULL) {
            if (i_dev != NULL) {
                if (!strcmp(i_dev, serial)) {
                    d = i;
                    continue;
                }
            } else if (!strcmp(desc, "IDE64 USB DEVICE")) {
                d = i;
                if (ad == NULL) continue;
                strcpy(desc, desc2);
                strcpy(serial, serial2);
            } else if (ad == NULL && !strcmp(desc, "FT245R USB FIFO")) {
                strcpy(desc2, desc);
                strcpy(serial2, serial);
                ad = i;
                continue;
            }
        }
        log_printf("Ignored device: %s (%s)", serial, desc);
    }
    if (d == NULL && ad != NULL) {
        d = ad;
        strcpy(desc, desc2);
        strcpy(serial, serial2);
    }
    if (d == NULL) {
        log_print("No usable USB device found for now");
        ftdi_list_free(&list);
        return 1;
    }
    if (ftdi_usb_open_dev(ftDevice, d->dev)) {
        log_printf("ftdi_usb_open_dev: %s", ftdi_get_error_string(ftDevice));
        ftdi_list_free(&list);
        return 1;
    }
    ftdi_list_free(&list);
    if (ftdi_usb_reset(ftDevice)) {
        log_printf("ftdi_usb_reset: %s", ftdi_get_error_string(ftDevice));
        ftdi_usb_close(ftDevice);
        return 1;
    }
    log_printf("Served device: %s (%s)", serial, desc);

    ftdi_usb_purge_buffers(ftDevice);
    ftdi_set_error_char(ftDevice, 0, 0);
    ftdi_read_data_set_chunksize(ftDevice, 512);
#ifdef LIBFTDI_LATENCY
    last_latency = -1;
#else
    ftdi_set_latency_timer(ftDevice, 4);
#endif
#else
    DWORD n, i, j;
    FT_DEVICE_LIST_INFO_NODE *list;
    FT_STATUS ftStatus;

    driver_errno = DRIVER_EIO;
    inited = 0;

    if (!ftd2xx_loaded) {
        log_print("FTD2XX USB driver not found");
        return 1;
    }

    ftStatus = FT_CreateDeviceInfoList(&n);
    if (ftStatus != FT_OK) {
        log_print("FT_CreateDeviceInfoList failed");
        return 1;
    }
    if (n == 0) {
        log_print("No usable USB device found for now");
        return 1;
    }

    list = (FT_DEVICE_LIST_INFO_NODE *)malloc(n * sizeof *list);
    if (!list) {
        log_print("Out of memory");
        return 1;
    }
    ftStatus = FT_GetDeviceInfoList(list, &n);
    if (ftStatus != FT_OK) n = 0;
    for (i = 0; i < n; i++) {
        if (i_dev != NULL) {
            if (!strcmp(i_dev, list[i].SerialNumber)) break;
        } else {
            if (!strcmp("IDE64 USB DEVICE", list[i].Description)) break;
        }
    }
    if (i == n && i_dev == NULL) {
        for (i = 0; i < n; i++) {
            if (!strcmp("FT245R USB FIFO", list[i].Description)) break;
        }
    }
    if (i == n) {
        free(list);
        log_print("No usable USB device found for now");
        return 1;
    }
    ftStatus = FT_OpenEx((void*)list[i].LocId, FT_OPEN_BY_LOCATION, &ftHandle);
    if (ftStatus != FT_OK) ftStatus = FT_OpenEx(list[i].SerialNumber, FT_OPEN_BY_SERIAL_NUMBER, &ftHandle);
    if (ftStatus != FT_OK) {
        free(list);
        ftHandle = NULL;
        log_print("FT_OpenEx failed");
        return 1;
    }
    for (j = 0; j < n; j++) {
        if (i == j) {
            continue;
        }
        log_printf("Ignored device: %s (%s)", list[j].SerialNumber, list[j].Description);
    }
    log_printf("Served device: %s (%s)", list[i].SerialNumber, list[i].Description);
    free(list);

    FT_ResetDevice(ftHandle);
    FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
    FT_SetTimeouts(ftHandle, 100, 10000);       // C64->PC, PC->C64
    FT_SetLatencyTimer(ftHandle, 1);
//     FT_SetUSBParameters(ftHandle, 64,0);
//     FT_SetDeadmanTimeout(ftHandle, 10);
#endif
    timeout_init();
    last_ec = -1;
    inited = 1;
    return 0;
}

static int getb(int use_timeout) {
    unsigned char data;
    if (driver_errno != DRIVER_SUCCESS) return EOF;
    if (use_timeout) timeout_set(1);
    do {
#ifndef FTD2XX_API
        int err = ftdi_read_data(ftDevice, &data, 1);
        if (err == -ENODEV) {
            driver_errno = DRIVER_ENODEV;
            return EOF;
        }
        if (err < 0) {
            driver_errno = DRIVER_EIO;
            return EOF;
        }
        if (err == 1) {
            return data;
        }
#else
        DWORD dwBytesRead;
        FT_STATUS ftStatus = FT_Read(ftHandle, &data, 1, &dwBytesRead);
        if (ftStatus == FT_DEVICE_NOT_FOUND) {
            driver_errno = DRIVER_ENODEV;
            return EOF;
        }
        if (ftStatus != FT_OK) {
            driver_errno = DRIVER_ENODEV;
            return EOF;
        }
        if (dwBytesRead == 1) {
            return data;
        }
#endif
    } while (!timeout_expired);
    driver_errno = DRIVER_EIO;
    return EOF;
}

static void getbytes(unsigned char data[], unsigned int bytes) {
    int err = 1;
#ifdef FTD2XX_API
    DWORD dwBytesRead;
    FT_STATUS ftStatus;
#endif
    if (driver_errno != DRIVER_SUCCESS) return;
    do {
        unsigned int l;
        if (err != 0) timeout_set(1);
        if (bytes > 4096) l = 4096; else l = bytes;
        if (timeout_expired) {
            driver_errno = DRIVER_EIO;
            return;
        }
#ifndef FTD2XX_API
        err = ftdi_read_data(ftDevice, data, l);
        if (err == -ENODEV) {
            driver_errno = DRIVER_ENODEV;
            return;
        }
        if (err < 0) {
            driver_errno = DRIVER_EIO;
            return;
        }
        l = (unsigned int)err;
        bytes -= l; data += l;
#else
        ftStatus = FT_Read(ftHandle, data, l, &dwBytesRead);
        if (ftStatus == FT_DEVICE_NOT_FOUND) {
            driver_errno = DRIVER_ENODEV;
            return;
        }
        if (ftStatus != FT_OK) {
            driver_errno = DRIVER_EIO;
            return;
        }
        bytes -= dwBytesRead; data += dwBytesRead;
        err = (int)dwBytesRead;
#endif
    } while (bytes > 0);
}

static void sendb(unsigned char data) {
    if (ebufop >= sizeof ebufo) {
        driver_errno = flush();
        if (driver_errno != DRIVER_SUCCESS) return;
    }
    ebufo[ebufop++] = data;
}

static void sendbytes(const unsigned char data[], unsigned int bytes) {
    while (ebufop + bytes > sizeof ebufo) {
        memcpy(ebufo + ebufop, data, sizeof(ebufo) - ebufop);
        data += sizeof(ebufo) - ebufop;
        bytes -= sizeof(ebufo) - ebufop;
        ebufop = sizeof ebufo;
        driver_errno = flush();
        if (driver_errno != DRIVER_SUCCESS) return;
    }
    memcpy(ebufo + ebufop, data, bytes);
    ebufop += bytes;
}

static void eshutdown(void) {
#ifndef FTD2XX_API
    if (ftDevice != NULL) {
        if (inited) {
            ftdi_usb_purge_buffers(ftDevice);
            ftdi_usb_close(ftDevice);
        }
        ftdi_free(ftDevice);
        ftDevice = NULL;
    }
#else
    if (ftHandle != NULL) {
        if (inited) {
            FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
        }
        FT_Close(ftHandle);
        ftHandle = NULL;
    }
#endif
#ifdef LIBUSB_HOTPLUG
    if (usb != NULL) {
        libusb_exit(usb);
        usb = NULL;
    }
#endif
    timeout_deinit();
    inited = 0;
}

static Driver_errno flush(void) {
    if (ebufop && driver_errno == DRIVER_SUCCESS) {
        int l = ebufop;
        unsigned char *data = ebufo;
        ebufop = 0;
        do {
#ifndef FTD2XX_API
            int err = ftdi_write_data(ftDevice, data, l);
            if (err == -ENODEV) return DRIVER_ENODEV;
            if (err < 0) return DRIVER_EIO;
            l -= err; data += err;
#else
            DWORD dwBytesWritten;
            FT_STATUS ftStatus = FT_Write(ftHandle, data, l, &dwBytesWritten);
            if (ftStatus == FT_DEVICE_NOT_FOUND) return DRIVER_ENODEV;
            if (ftStatus != FT_OK) return DRIVER_EIO;
            l -= dwBytesWritten; data += dwBytesWritten;
#endif
        } while (l > 0);
    }
    return driver_errno;
}

static Driver_errno done(void) {
    timeout_cancel();
    return driver_errno;
}

static void turn(int m) {
#ifdef LIBFTDI_LATENCY
    int latency = m < 0 ? 255 : 1;
    if (last_latency != latency) {
        ftdi_set_latency_timer(ftDevice, latency);
        last_latency = latency;
    }
#endif
    if (last_ec != m) {
#ifndef FTD2XX_API
        ftdi_set_event_char(ftDevice, m, m >= 0);
#else
        FT_SetChars(ftHandle, m, m >= 0, 0, 0);
#endif
        last_ec = m;
    }
}

static int clean(void) {
#ifndef FTD2XX_API
    ftdi_usb_purge_buffers(ftDevice);
#else
    FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
#endif
    ebufop = 0;
    return timeout_expired;
}

static int waitb(unsigned char ec) {
#ifndef FTD2XX_API
#ifndef LIBFTDI_LATENCY
    unsigned int dyntime = 1;
#endif
#endif
    unsigned char data;
    if (!inited) return DRIVER_ENODEV;
    timeout_cancel();
    if (ec != last_ec) {
#ifndef FTD2XX_API
        ftdi_set_event_char(ftDevice, ec, 1);
#else
        FT_SetChars(ftHandle, ec, 1, 0, 0);
#endif
        last_ec = ec;
    }
#ifdef LIBFTDI_LATENCY
    if (last_latency != 255) {
        ftdi_set_latency_timer(ftDevice, 255);
        last_latency = 255;
    }
#endif
    while (!signal_received) {
#ifndef FTD2XX_API
        int err = ftdi_read_data(ftDevice, &data, 1);
        if (err == -ENODEV) return DRIVER_ENODEV;
        if (err < 0) return DRIVER_EIO;
        if (err == 1) {
#ifdef LIBFTDI_LATENCY
            turn(ec);
#endif
            break;
        }
#ifndef LIBFTDI_LATENCY
        usleep(dyntime);
        if (dyntime < 16384) dyntime <<= 1;
#endif
#else
        DWORD dwBytesRead;
        FT_STATUS ftStatus = FT_Read(ftHandle, &data, 1, &dwBytesRead);
        if (ftStatus == FT_DEVICE_NOT_FOUND) return DRIVER_ENODEV;
        if (ftStatus != FT_OK) return DRIVER_ENODEV;
        if (dwBytesRead == 1) {
            break;
        }
#endif
    }
    driver_errno = DRIVER_SUCCESS; timeout_expired = 0;
    return signal_received ? DRIVER_EINTR : data;
}

static void destroy(void) {
#ifdef FTD2XX_API
    if (ftd2xx_loaded) {
        ftd2xx_deinit();
        ftd2xx_loaded = 0;
    }
#endif
}

static const Driver driver = {
    .mode         = M_USB,
    .name         = "USB",
    .initialize   = initialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .done         = done,
    .turn         = turn,
    .wait         = waitb,
    .clean        = clean,
    .destroy      = destroy,
};

const Driver *usb_driver(const char *dev) {
    i_dev = dev;
#ifdef FTD2XX_API
    if (!ftd2xx_loaded) {
        ftd2xx_loaded = ftd2xx_init();
    }
#endif
    log_printf("Using %s driver for device %s", driver.name, dev != NULL ? dev : "*");
    return &driver;
}
#else
const Driver *usb_driver(const char *dev) {
    (void)dev;
    log_print("USB libraries were not available at compile time");
    return NULL;
}
#endif
